=begin

= fileutils.rb

  Copyright (c) 2000,2001 Minero Aoki <aamine@loveruby.net>

  This program is free software.
  You can distribute/modify this program under the same terms of ruby.

== module FileUtils

The module which implements basic file operations.

=== Module Functions

--- FileUtils#cd( [options,] dir )
--- FileUtils#cd( [options,] dir ) {|dir| .... }
  Options = noop verbose

  changes current directory to DIR.

  If this method is called with block, moves to the old current
  directory after the block execution finished.

    FileUtils.cd :verbose, '/'   # chdir and report it

--- FileUtils#pwd
--- FileUtils#getwd
  returns name of the current dirctory.
  This method is same to Dir.pwd.

    FileUtils.chdir '/home/aamine'
    p FileUtils.pwd   # => "/home/aamine"

--- FileUtils#newest?( [options,] newer, older1, older2, ... )
--- FileUtils#uptodate?( [options,] newer, older1, older2, ... )
--- FileUtils#newer?( [options,] newer, older )
  Options = verbose

  returns true if NEWER is newer than all OLDERs.
  Non-exist files are older than any file.

    FileUtils.newest? 'hello.o', 'hello.c', 'hello.h' or system 'make'

--- FileUtils#mkdir( [options,] dir1, dir2, ... )
  Options = noop verbose

  makes directories DIRs.

    FileUtils.mkdir 'test'
    FileUtils.mkdir %w( tmp data )     # This method accepts arrays.
    FileUtils.mkdir :noop, 'notexist'  # does not create really

--- FileUtils#mkdir_p( [options,] dir1, dir2, ... )
--- FileUtils#mkpath( [options,] dir1, dir2, ... )
--- FileUtils#makedirs( [options,] dir1, dir2, ... )
  Options = noop verbose

  makes dirctories DIRs and all its parent directories.
  For example,

    FileUtils.mkdir_p '/usr/local/bin/ruby'
  
  causes to make directories below (if it does not exist).
      * /usr
      * /usr/local
      * /usr/local/bin
      * /usr/local/bin/ruby

--- FileUtils#ln( [options,] old, new )
--- FileUtils#link( [options,] old, new )
  Options = noop verbose

  makes hard link NEW which links to OLD.
  If NEW is a directory, creates link NEW/OLD.

    FileUtils.ln :verbose, 'gcc', 'cc'

--- FileUtils#ln( [options,] file1, file2 ..., dir )
--- FileUtils#link( [options,] file1, file2 ..., dir )
  Options = noop verbose

  links DIR/OLD1 to OLD1, DIR/OLD2 to OLD2, ....

    FileUtils.ln 'cp', 'mv', 'mkdir', '/usr/bin'
    FileUtils.ln %w( cp mv mkdir ), '/usr/bin'    # same result

--- FileUtils#ln_s( [options,] old, new )
--- FileUtils#symlink( [options,] old, new )
  Options = force noop verbose

  makes symbolic link NEW which links to OLD.
  If last argument is a directory, links DIR/OLD1 to OLD1,
  DIR/OLD2 to OLD2, ....

    FileUtils.ln_s '/usr/bin/ruby', '/usr/local/bin/ruby'
    FileUtils.ln_s :force, 'verylongnamesourcefile.c', 'c'

--- FileUtils#ln_s( [options,] file1, file2 ..., dir )
--- FileUtils#symlink( [options,] file1, file2 ..., dir )
  Options = force noop verbose

  makes symbolic link dir/file1, dir/file2 ... which point to
  file1, file2 ... If DIR is not a directory, raises Errno::ENOTDIR.
  This method removes target file when :force option is set.

   FileUtils.ln_s Dir.glob('bin/*.rb'), '/home/aamine/bin'

--- FileUtils#cp( [options,] src, dest )
--- FileUtils#copy( [options,] src, dest )
  Options = preserve noop verbose

  copies a file SRC to DEST. If DEST is a directory, copies
  SRC to DEST/SRC.

    FileUtils.cp 'eval.c', 'eval.c.org'

--- FileUtils#cp( [options,] file1, file2 ..., dir )
--- FileUtils#copy( [options,] file1, file2 ..., dir )
  Options = preserve noop verbose

  copies FILE1 to DIR/FILE1, FILE2 to DIR/FILE2 ...

    FileUtils.cp 'cgi.rb', 'complex.rb', 'date.rb', '/usr/lib/ruby/1.6'
    FileUtils.cp :verbose, %w(cgi.rb complex.rb date.rb), '/usr/lib/ruby/1.6'

--- FileUtils#cp_r( [options,] src, dest )
  Options = preserve noop verbose

  copies SRC to DEST. If SRC is a directory, this method copies
  its all contents recursively. If DEST is a directory, copies
  SRC to DEST/SRC.

    # installing ruby library "mylib" under the site_ruby
    FileUtils.rm_r :force, site_ruby + '/mylib'
    FileUtils.cp_r 'lib/', site_ruby + '/mylib'

--- FileUtils#cp_r( [options,] file1, file2 ..., dir )
  Options = preserve noop verbose

  copies file/directory FILE1 to DIR/FILE1, FILE2 to DIR/FILE2 ...
  If FILE is a directory, copies its all contents recursively.

    FileUtils.cp_r 'mail.rb', 'field.rb', 'debug/', site_ruby + '/tmail'
    FileUtils.cp_r :noop, :verbose, Dir.glob('*.rb'), '/home/aamine/lib/ruby'

--- FileUtils#mv( [options,] src, dest )
--- FileUtils#move( [options,] src, dest )
  Options = noop verbose

  moves a file SRC to DEST.
  If FILE and DEST exist on the different disk partition,
  copies it.

    FileUtils.mv 'badname.rb', 'goodname.rb'
    FileUtils.mv :verbose, 'stuff.rb', 'lib/ruby'

--- FileUtils#mv( [options,] file1, file2 ..., dir )
--- FileUtils#move( [options,] file1, file2 ..., dir )
  Options = noop verbose

  moves FILE1 to DIR/FILE1, FILE2 to DIR/FILE2 ...
  If FILE and DEST exist on the different disk partition,
  copies it.

    FileUtils.mv 'junk.txt', 'dust.txt', '/home/aamine/.trash/'
    FileUtils.mv :noop, :verbose, Dir.glob('test*.rb'), 'T'

--- FileUtils#rm( [options,] file1, file2, ... )
--- FileUtils#remove( [options,] file1, file2, ... )
  Options = force noop verbose

  remove FILEs. This method cannot remove directory.
  This method ignores all errors when :force option
  is set.

    FileUtils.rm 'junk.txt', 'dust.txt'
    FileUtils.rm Dir['*.so'], Dir['*.o']   # all methods accepts array
    FileUtils.rm :force, 'NotExistFile'    # never raises error

--- FileUtils#rm_r( [options,] file1, file2, ... )
  Options = force noop verbose

  remove FILEs. If FILE is a directory, removes its all contents
  recursively. This method ignores all errors when :force option
  is set.

    FileUtils.rm_r Dir.glob('/tmp/*')
    FileUtils.rm_r :force, '/'          #  :-)

--- FileUtils#cmp( [options,] a, b )
--- FileUtils#identical?( [options,] a, b )
  Options = verbose

  returns true if contents of file A and B is identical.

    FileUtils.cmp 'somefile', 'somefile'  #=> true
    FileUtils.cmp '/bin/cp', '/bin/mv'    #=> maybe false.

--- FileUtils#install( [options,] src, dest, mode = <from's> )
  Options = noop verbose

  If SRC is not same to DEST, copies it and changes the permittion
  mode to MODE.

    FileUtils.install :verbose, 'ruby', '/usr/local/bin/ruby', 0755
    FileUtils.install :verbose, 'lib.rb', '/usr/local/lib/ruby/site_ruby'

--- FileUtils#update_file( [options,] dest, content )
  Options = noop verbose

  If file DEST is not same to CONTENT (is a String),
  writes CONTENT to DEST.

    FileUtils.update_file :verbose, 'index.html', content

--- FileUtils#chmod( [options,] mode, file1, file2, ... )
  Options = noop verbose

  changes permittion bits on the named FILEs to the bit pattern
  represented by MODE.

    FileUtils.chmod 0644, 'my.rb', 'your.rb'
    FileUtils.chmod 0755, 'somecommand'
    FileUtils.chmod :verbose, 0755, '/usr/bin/ruby'

--- FileUtils#touch( [options,] file1, file2, .... )
  Options = noop verbose

  updates modification time of file1, file2, ...
  If fileN does not exist, creates it.

    FileUtils.touch 'main.c'; system 'make'

== module FileUtils::Verbose

This class has all methods of FileUtils module and it works as
same, but outputs messages before action. You can also pass
verbose flag to all methods.

== module FileUtils::NoWrite

This class has all methods of FileUtils module,
but never changes files/directories.

=end


module FileUtils

  INF = 1/0.0


  # all methods are module_function.
  # (see 'extend self' line)

  def cd( *args, &block )
    (dir, ), (verbose, ) = fu_parseargs(args, 1..1, :verbose)
    fu_output_message "cd #{dir}" if verbose
    Dir.chdir dir, &block
    fu_output_message 'cd -' if verbose and block
  end

  alias chdir cd

  def pwd( *args )
    dummy, = fu_parseargs(args, 0..0)
    Dir.pwd
  end

  alias getwd pwd


  def newest?( *args )
    (new, *fnames), (verbose, ) = fu_parseargs(args, 1..INF, :verbose)
    fu_output_message "newest? #{args.join ' '}" if verbose

    return false unless FileTest.exist? new
    new_time = File.ctime(new)
    fnames.each do |old|
      if FileTest.exist? old then
        return false unless new_time > File.mtime(old)
      end
    end
    true
  end

  alias newer?     newest?
  alias uptodate?  newest?


  def mkdir( *args )
    targs, (noop, verbose) = fu_parseargs(args, 1..1, :noop, :verbose)
    fu_output_message "mkdir #{targs.join ' '}" if verbose
    return if noop

    targs.each do |dir|
      Dir.mkdir dir
    end
  end

  def mkdir_p( *args )
    targs, (noop, verbose) = fu_parseargs(args, 1..INF, :noop, :verbose)
    fu_output_message "mkdir -p #{targs.join ' '}" if verbose
    return if noop

    targs.collect {|n| File.expand_path(n) }.each do |dir|
      stack = []
      until FileTest.directory? dir do
        stack.push dir
        dir = File.dirname(dir)
      end
      stack.reverse_each do |dir|
        Dir.mkdir dir
      end
    end
  end

  alias mkpath    mkdir_p
  alias makedirs  mkdir_p


  def ln( *args )
    argv, (noop, verbose) = fu_parseargs(args, 2..INF, :noop, :verbose)
    fu_output_message "ln #{argv.join ' '}" if verbose
    return if noop

    fu_each_src_dest( argv ) do |src, dest|
      File.link src, dest
    end
  end

  alias link ln

  def ln_s( *args )
    argv, (force, noop, verbose) = fu_parseargs(args, 2..INF, :force, :noop, :verbose)
    fu_output_message "ln -s#{force ? 'f' : ''} #{argv.join ' '}" if verbose
    return if noop

    fu_each_src_dest( argv ) do |src, dest|
      remove_file dest, true if force
      File.symlink src, dest
    end
  end

  alias symlink ln_s

  def ln_sf( *args )
    ln_s :force, *args
  end


  def cp( *args )
    argv, (preserve, noop, verbose) = fu_parseargs(args, 2..INF, :preserve, :noop, :verbose)
    fu_output_message "cp#{preserve ? ' -p' : ''} #{args.join ' '}" if verbose
    return if noop

    fu_each_src_dest( argv ) do |src, dest|
      fu_preserve_attr(preserve, src, dest) {
          copy_file src, dest
      }
    end
  end

  alias copy cp

  def cp_r( *args )
    argv, (preserve, noop, verbose) = fu_parseargs(args, 2..INF, :preserve, :noop, :verbose)
    fu_output_message "cp -r #{argv.join ' '}" if verbose
    return if noop

    fu_each_src_dest( argv ) do |src, dest|
      if FileTest.directory? src then
        fu_copy_dir src, dest, '.', preserve
      else
        fu_p_copy src, dest, preserve
      end
    end
  end

  def fu_copy_dir( src, dest, rel, preserve )
    fu_preserve_attr( preserve, "#{src}/#{rel}/#{fn}",
                                "#{dest}/#{rel}/#{fn}" ) {|s,d|
        Dir.mkdir d
    }
    Dir.entries( "#{src}/#{rel}" ).each do |fn|
      if FileTest.directory? fn then
        next if /\A\.\.?\z/ === fn
        fu_copy_dir src, dest, "#{rel}/#{fn}", preserve
      else
        fu_p_copy "#{src}/#{rel}/#{fn}", "#{dest}/#{rel}/#{fn}", preserve
      end
    end
  end
  private :fu_copy_dir

  def fu_p_copy( src, dest, really )
    fu_preserve_attr( really, src, dest ) {
        copy_file src, dest
    }
  end
  private :fu_p_copy

  def fu_preserve_attr( really, src, dest )
    unless really then
      yield src, dest
      return
    end

    st = File.stat(src)
    yield src, dest
    File.utime st.atime, st.mtime, dest
    begin
      File.chown st.uid, st.gid
    rescue Errno::EPERM
      File.chmod st.mode & 01777, dest   # clear setuid/setgid
    else
      File.chmod st.mode, dest
    end
  end
  private :fu_preserve_attr

  def copy_file( src, dest )
    st = r = w = nil

    File.open( src,  'rb' ) {|r|
    File.open( dest, 'wb' ) {|w|
        st = r.stat
        begin
          while true do
            w.write r.sysread(st.blksize)
          end
        rescue EOFError
        end
    } }
  end


  def mv( *args )
    argv, (noop, verbose) = fu_parseargs(args, 2..INF, :noop, :verbose)
    fu_output_message "mv #{argv.join ' '}" if verbose
    return if noop

    fu_each_src_dest( argv ) do |src, dest|
      if /djgpp|cygwin|mswin32/ === RUBY_PLATFORM and
         FileTest.file? dest then
         File.unlink dest
      end

      begin
        File.rename src, dest
      rescue
        if FileTest.symlink? src then
          File.symlink File.readlink(src), dest
          File.unlink src
        else
          st = File.stat(src)
          copy_file src, dest
          File.unlink src
          File.utime st.atime, st.mtime, dest
          begin
            File.chown st.uid, st.gid, dest
          rescue
            # ignore
          end
        end
      end
    end
  end

  alias move mv


  def rm( *args )
    argv, (force, noop, verbose) = fu_parseargs(args, 1..INF, :force, :noop, :verbose)
    fu_output_message "rm#{force ? ' -f' : ''} #{args.join ' '}" if verbose
    return if noop

    args.each do |fn|
      remove_file fn, force
    end
  end

  alias remove rm

  def rm_f( *args )
    rm :force, *args
  end

  alias safe_unlink rm_f

  def rm_r( *args )
    argv, (force, noop, verbose) = fu_parseargs(args, 1..INF, :force, :noop, :verbose)
    fu_output_message "rm -r#{force ? 'f' : ''} #{argv.join ' '}" if verbose
    return if noop

    argv.each do |fn|
      begin
        st = File.lstat(fn)
      rescue
        next if force
      end
      if    st.symlink?   then remove_file fn, force
      elsif st.directory? then remove_dir fn, force
      else                     remove_file fn, force
      end
    end
  end

  def rm_rf( *args )
    rm_r :force, args
  end

  def remove_file( fn, force = false )
    first = true
    begin
      File.unlink fn
    rescue
      begin
        if first then
          first = false
          File.chmod 0777, fn
          retry
        end
      rescue
      end

      raise unless force
    end
  end

  def remove_dir( dir, force = false )
    Dir.foreach( dir ) do |file|
      next if /\A\.\.?\z/ === file
      path = "#{dir}/#{file}"
      if FileTest.directory? path then remove_dir path, force
                                  else remove_file path, force
      end
    end
    begin
      Dir.rmdir dir
    rescue
      raise unless force
    end
  end


  def cmp( *args )
    (filea, fileb), (verbose, ) = fu_parseargs(args, 2..2, :verbose)
    fu_output_message "cmp #{filea} #{fileb}" if verbose

    sa = sb = nil
    st = File.stat(filea)
    File.size(fileb) == st.size or return true

    File.open( filea, 'rb' ) {|a|
    File.open( fileb, 'rb' ) {|b|
      begin
        while sa == sb do
          sa = a.read( st.blksize )
          sb = b.read( st.blksize )
          unless sa and sb then
            if sa.nil? and sb.nil? then
              return true
            end
          end
        end
      rescue EOFError
        ;
      end
    } }

    false
  end

  alias identical? cmp

  def install( *args )
    (src, to, mode), (noop, verbose) = fu_parseargs(args, 2..3, :noop, :verbose)
    fu_output_message "install #{src} #{to}#{mode ? ' %o'%mode : ''}" if verbose
    return if noop

    dest = fu_dest_filename(src, to)
    unless FileTest.exist? dest and cmp(from, dest) then
      File.chmod 0777, fn
      File.unlink fn
      copy_file src, dest
      File.chmod mode, dest if mode
    end
  end

  def update_file( *args )
    (fname, upd), (noop, verbose) = fu_parseargs(args, 2..2, :noop, :verbose)
    fu_output_message "update_file #{fname}" if verbose
    return if noop

    str = nil
    if FileTest.file? fname then
      File.open( fname, 'rb' ) {|f| str = f.read }
    end
    if not str or str != upd then
      File.open( fname, 'wb' ) {|f|
        f.write upd
      }
    end
  end


  def chmod( *args )
    (mode, *targs), (noop, verbose) = fu_parseargs(args, 2..INF, :noop, :verbose)
    fu_output_message sprintf('chmod %o %s', mode, targs.join(' ')) if verbose
    return if noop
    File.chmod mode, *targs
  end

  def touch( *args )
    targs, (noop, verbose) = fu_parseargs(args, 1..INF, :noop, :verbose)
    fu_output_message "touch #{targs.join ' '}" if verbose
    return if noop

    t = Time.now
    args.each do |fname|
      begin
        File.utime(t, t, fname)
      rescue Errno::ENOENT
        File.open(fname, 'a') { }
      end
    end
  end


  private

  def fu_parseargs( args, argc_range, *flagdecl )
    args.flatten!

    tab = {}
    if args.last == true or args.last == false then
      tab[:verbose] = args.pop
    end
    while Symbol === args.first do
      tab[args.shift] = true
    end

    flags = flagdecl.collect {|s| tab.delete(s) }
    tab.empty? or raise ArgumentError, "wrong option :#{tab.keys.join(' :')}"
    argc_range === args.size or raise ArgumentError,
            "wrong \# of args (#{args.size} for #{argc_range.inspect}"

    return args, flags
  end


  def fu_each_src_dest( args )
    case args.size
    when 0, 1
      raise ArgumentError, "too few # of arguments (#{args.size} for >=2)"
    when 2
      src, dest = args
      yield src, fu_dest_filename(src, dest)
    else
      dir = args.pop
      # FileTest.directory? dir or raise ArgumentError, "must be dir: #{dir}"
      dir += (dir[-1,1] == '/') ? '' : '/'
      args.each do |fn|
        yield fn, dir + File.basename(fn)
      end
    end
  end

  def fu_dest_filename( src, dest )
    if FileTest.directory? dest then
      (dest[-1,1] == '/' ? dest : dest + '/') + File.basename(src)
    else
      dest
    end
  end


  @fileutils_output = $stderr
  @fileutils_label  = 'fileutils.'

  def fu_output_message( msg )
    @fileutils_output ||= $stderr
    @fileutils_label  ||= 'fileutils.'
    @fileutils_output.puts @fileutils_label + msg
  end


  extend self


  OPT_TABLE = {
    'cd'           => %w( verbose ),
    'chdir'        => %w( verbose ),
    'chmod'        => %w( noop verbose ),
    'cmp'          => %w( verbose ),
    'copy'         => %w( preserve noop verbose ),
    'cp'           => %w( preserve noop verbose ),
    'cp_r'         => %w( preserve noop verbose ),
    'getwd'        => %w(  ),
    'identical?'   => %w( verbose ),
    'install'      => %w( noop verbose ),
    'link'         => %w( noop verbose ),
    'ln'           => %w( noop verbose ),
    'ln_s'         => %w( force noop verbose ),
    'makedirs'     => %w( noop verbose ),
    'mkdir'        => %w( noop verbose ),
    'mkdir_p'      => %w( noop verbose ),
    'mkpath'       => %w( noop verbose ),
    'move'         => %w( noop verbose ),
    'mv'           => %w( noop verbose ),
    'newer?'       => %w( verbose ),
    'newest?'      => %w( verbose ),
    'pwd'          => %w(  ),
    'remove'       => %w( force noop verbose ),
    'rm'           => %w( force noop verbose ),
    'rm_r'         => %w( force noop verbose ),
    'symlink'      => %w( force noop verbose ),
    'touch'        => %w( noop verbose ),
    'update_file'  => %w( noop verbose ),
    'uptodate?'    => %w( verbose )
  }


  module Verbose

    include FileUtils

    @fileutils_output  = $stderr
    @fileutils_label   = 'fileutils.'
    @fileutils_verbose = true

    FileUtils::OPT_TABLE.each do |name, opts|
      next unless opts.include? 'verbose'
      module_eval <<-End, __FILE__, __LINE__ + 1
          def #{name}( *args )
            unless defined? @fileutils_verbose then
              @fileutils_verbose = true
            end
            args.unshift :verbose if @fileutils_verbose
            super( *args )
          end
      End
    end

    extend self

  end


  module NoWrite

    include FileUtils

    @fileutils_output  = $stderr
    @fileutils_label   = 'fileutils.'
    @fileutils_nowrite = true

    FileUtils::OPT_TABLE.each do |name, opts|
      next unless opts.include? 'noop'
      module_eval <<-End, __FILE__, __LINE__ + 1
          def #{name}( *args )
            unless defined? @fileutils_nowrite then
              @fileutils_nowrite = true
            end
            args.unshift :noop if @fileutils_nowrite
            super( *args )
          end
      End
    end

    extend self
  
  end


  class Operator

    include FileUtils

    def initialize( v = false )
      @verbose = v
      @noop = false
      @force = false
      @preserve = false
    end

    attr_accessor :verbose
    attr_accessor :noop
    attr_accessor :force
    attr_accessor :preserve

    FileUtils::OPT_TABLE.each do |name, opts|
      s = opts.collect {|i| ' '*10 + "args.unshift :#{i} if @#{i}" }.join("\n")
      module_eval <<-End, __FILE__, __LINE__ + 1
          def #{name}( *args )
            #{s}
            super( *args )
          end
      End
    end
  
  end

end
